using Kafka.Public;
using Kafka.Public.Loggers;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;

namespace franz
{
    public class Telefranz
    {
        static class _TelefranzConsumers<T> where T : silver_messages.message, new() {
            internal static KafkaConsumer<string, T> consumer = null;
            internal static readonly Dictionary<Action<T>, Action<KafkaRecord<string, T>>> wrappings 
                = new Dictionary<Action<T>, Action<KafkaRecord<string, T>>>();
            internal static uint references = 0;
        }
        private string handling_group { get; set; } = "Liszt";
        const string namespace_prefix = "silver_messages";
        private silver_messages.global.report howToReport { get; set; }
        private bool shouldListen = false;
        private ClusterClient clusterClient;
        private Task listenTask = null;
        private List<object> listeningConsumers = new List<object>();

        public Telefranz(string name, string bootstrap_servers,
            List<string> commands = null, List<string> checks = null,
            List<string> errors = null, List<string> warnings = null)
        {
            clusterClient = new ClusterClient(new Kafka.Public.Configuration { Seeds = bootstrap_servers, ClientId = name }, new ConsoleLogger());

            this.handling_group = name;
            this.howToReport = new silver_messages.global.report()
            {
                name = name,
                errors = errors ?? new List<string>(),
                warnings = warnings ?? new List<string>()
            };
            if (commands?.Count > 0)
            {
                howToReport.capabilites.commands = commands;
            }
            if (checks?.Count > 0)
            {
                howToReport.capabilites.checks = checks;
            }

            addHandler<silver_messages.global.sound_off>((m) =>
            {
                ProduceMessage(this.howToReport);
            });
            addHandler<silver_messages.global.stop>((m) =>
            {
                if (m.name == this.handling_group)
                {
                    Environment.Exit(0);
                }
            });
            addHandler<silver_messages.global.restart>((m) =>
            {
                if (m.name == this.handling_group)
                {
                    throw new NotImplementedException();//TODO
                }
            });
        }

        public void addHandler<T>(Action<T> theAction) where T : silver_messages.message, new()
        {
            var dummy = new T();
            if(_TelefranzConsumers<T>.consumer == null)
            {
                _TelefranzConsumers<T>.consumer = new KafkaConsumer<string, T>(dummy.topic, clusterClient);
            }
            
            Action<KafkaRecord<string, T>> wrapped = (kr) =>
            {
                theAction(kr.Value);
            };
            _TelefranzConsumers<T>.consumer.MessageReceived += wrapped;
            _TelefranzConsumers<T>.wrappings[theAction] = wrapped;
            _TelefranzConsumers<T>.references++;
            if(_TelefranzConsumers<T>.references == 1)
            {
                _TelefranzConsumers<T>.consumer.Subscribe(this.handling_group, 
                    new List<string>(){dummy.topic}, 
                    new ConsumerGroupConfiguration());
                listeningConsumers.Add(_TelefranzConsumers<T>.consumer);
                _TelefranzConsumers<T>.consumer.ConsumeFromLatest();
            }
        }
        public void removeHandler<T>(Action<T> theAction) where T : silver_messages.message, new()
        {
            _TelefranzConsumers<T>.consumer.MessageReceived -= _TelefranzConsumers<T>.wrappings[theAction];
            _TelefranzConsumers<T>.wrappings.Remove(theAction);
            _TelefranzConsumers<T>.references--;
            if(_TelefranzConsumers<T>.references == 0)
            {
                _TelefranzConsumers<T>.consumer.StopConsume();
            }
        }

        public void ProduceMessage<T>(T message) where T : silver_messages.message
        {
            clusterClient.Produce(message.topic, message);
        }
    }
}